/*****************************************************************************
	    This file ("main.cpp") is a part of the webchat project.

This program is released, distributed and run in the terms of the 
General Public License

Copyright(c) 1999 Paul Fleischer
******************************************************************************/

#include "main.h"

//This points to the environment variable, containing the script's name
char *scriptName;

int
main(int args, char **argv)
{
  char *contentString;
  char *tmppoint;
  char *pathInfo;
  char *queryString;
  int contentLength;
  int contentSize;
  char key[2];

  scriptName = getenv("SCRIPT_NAME");

  printf("Content-Type: text/html\n\n");

  printf("<HTML>\n");

  pathInfo = getenv("PATH_INFO");
  tmppoint = getenv("CONTENT_LENGTH");
  queryString = getenv("QUERY_STRING");

  if(tmppoint != NULL)
    contentLength = atoi(tmppoint);
  else
    contentLength = 0;

  //Read POST-METHOD arguments
  if(contentLength > 0)
    {
      contentString = (char*) malloc(contentLength);
      fread(contentString, 1, contentLength, stdin);
    }

  //Show the login screen, if the script is called without any arguments
  if((contentLength < 1) && (pathInfo == NULL))
    loginScreen();

  if(pathInfo == NULL)
    {
      if(contentLength > 1)
	{
	  if( GetVar(contentString, "Key", key) )
	    {
	      if(key[0] == '1')
		login(contentString);
	      if(key[0] == '2')
		regNewUserForm();
	      if(key[0] == '3')
		regNewUser(contentString);
	    }
	}
    }
  else
    {
      if(strcmp(pathInfo, "/messages") == 0)
	showMessages(queryString);
      if(strcmp(pathInfo, "/send") == 0)
	{	
	  sendMessages(queryString);
	}
      if(strcmp(pathInfo, "/logout") == 0)
	{
	  char username[255];
	  char password[255];

	  GetVar(queryString, "Username", username);
	  GetVar(queryString, "Key", password);

	  if(!checkUserPass(true, username, password, NULL))
	    {
	      printf("Illegal access detected!");
	    }
	  else
	    {
	      logout(username);
	      printf("<HEAD><TITLE>Webchat %s: logged out</TITLE></HEAD>", VERSION);
	      printf("<BODY>");
	      printf("<H1>User '%s' logged out!</H1>\n", username);
	      printf("You can return to the <A href=\"%s\">start page</A>, if you want to log into another room<BR>\n", scriptName);
	    }
	}
    }

  //Footer... I know it's ugly ;)
  printf("<BR><BR><H5>Webchat script by <A href=mailto:polle@denmark.crosswinds.net>Paul Fleischer</A><BR>By request from <A href=http://www.coolweb.dk>www.coolweb.dk</A><BR><BR>Use of this script is on own risk!</BODY></HTML>");

  return 0;
}

/******************************************************************************
Shows the login screen.
******************************************************************************/
void
loginScreen(void)
{
  printf("<HEAD><TITLE>Webchat %s: login</TITLE></HEAD>\n", VERSION);
  printf("<BODY>");
  
  printf("<CENTER><H1>Login</H1>");
  printf("Existing user:<BR>\n");
  printf("<FORM action=\"%s\" method=post>\n", scriptName);
  printf("Username:<INPUT type=text name=Username><BR>\n");
  printf("Password:<INPUT type=password name=Password><BR>\n");
  printf("Room: <SELECT name=Room>\n");
  FILE *listfile;

  //Generate list of rooms, listed in ROOMLIST (defined in main.h)
  listfile = fopen(ROOMLIST, "r");
  if(listfile != NULL)
    {
      char line[255];
      while(!feof(listfile))
	{
	  readLine(listfile, line);
	  printf("<OPTION value=\"%s\">%s\n", line, line);
	}
      fclose(listfile);
    }
  else
    printf("<H1>No room list, chat will probably not work!</H1>\n");


  printf("</SELECT><BR>\n");
  printf("<INPUT type=hidden name=Key value=1>\n");
  printf("<INPUT type=submit value=Login><BR>\n");
  printf("</FORM>\n");
  printf("<HR>\n");
  printf("<FORM action=\"%s\" method=post>\n", scriptName);
  printf("<INPUT type=hidden name=Key value=2><BR>\n");
  printf("<INPUT type=submit value=\"Create a new user\"><BR>\n");
  printf("</FORM>\n");
  printf("<B>(Enter on your own risk, the programmer of this chatscript guarantess NOTHING!!!)</B>");
  printf("</CENTER>\n");
}

/******************************************************************************
Shows the "Register New User" screen
 *****************************************************************************/
void
regNewUserForm(void)
{
  printf("<HEAD><TITLE>Webchat %s: Registering new user</TITLE></HEAD>\n", VERSION);
  printf("<BODY>\n");

  printf("<CENTER>");
  printf("Fields marked with an '*' are <B>required</B><BR>");
  printf("<FORM action=\"%s\" method=post>\n", scriptName);
  printf("Username (*):<INPUT type=text name=Username><BR>\n");
  printf("E-mail:<INPUT type=text name=Email><BR>\n");
  printf("Password (*):<INPUT type=password name=Password><BR>\n");
  printf("Re-enter Password (*):<INPUT type=password name=Password1><BR>\n");
  printf("<INPUT type=hidden name=Key value=3>\n");
  printf("<INPUT type=submit value=Register><BR>\n");
  printf("</FORM>");
  printf("</CENTER>");
}

/******************************************************************************
Do the actual registering of a new user
 *****************************************************************************/
void
regNewUser(char *contentString)
{
  char username[255];
  char email[255];
  char password[255];
  char password1[255];
  bool ok=true;

  printf("<HEAD><TITLE>Webchat %s: Registering new user</TITLE></HEAD>\n", VERSION);

  if(!GetVar(contentString, "Username", username)) ok=false;
  GetVar(contentString, "Email", email);
  if(!GetVar(contentString, "Password", password)) ok=false;
  if(!GetVar(contentString, "Password1", password1)) ok=false;

  if(!ok)
    {
      printf("<H1>ERROR: Not all required fields have been filled</H1>\n");
      printf("Please fill out <B>all</B> required fields.<BR>\n");
      printf("(Press 'back' in your browser and try again");
      return;
    }

  if(strcmp(password, password1) != 0)
    {
      printf("<H1>ERROR: Passwords do not match</H1>\n");
      printf("The password typed in the 'Password' field, does not match with the one in the 'Re-enter Password' field!<BR>\n");
      printf("(Press 'back' in your browser and try again)");
    }

  if(checkUser(username))
    {
      printf("<H1>ERROR: User exists!</H1>");
      printf("The username '%s' has allready been registered, choose another one<BR>\n", username);
      printf("(Press 'back' in your browser and try again)");
      return;
    }

  if(addUser(username, email, password))
    {
      printf("<H1>User '%s' successful registered</H1>", username);
      printf("You can now go back to the <A href=\"%s\">start screen</A> and log into the chat", scriptName);
    }
  else
    {
      printf("<H1>Internal Script ERROR</H1>\n");
      printf("This script has encountered an internal script error<BR>\n");
      printf("The error code is <B>100-1</B><BR>\n");
      printf("Contact the webadmin of this website, and tell him about the error<BR>\n");
    }
}

/******************************************************************************
Adds a new user to the userdatabase, USERFILE (defined in main.h)
 *****************************************************************************/
bool
addUser(char *username, char *email, char *password)
{
  FILE *userfile;
  char salt[]="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
  char s[3];

  s[2] = 0;
  
  userfile = fopen(USERFILE, "a");
  if(userfile == NULL)
    return false;
  
  srand(time(NULL)*getpid());
  
  s[0] = salt[1+(int) (63.0*rand()/(RAND_MAX+1.0))];  
  s[1] = salt[1+(int) (63.0*rand()/(RAND_MAX+1.0))];
  
  fprintf(userfile, "%s:%s:%s\n", username, crypt(password, s), email);
  
  fclose(userfile);
  return true;
}

/******************************************************************************
Checks if the user 'username' exists
 *****************************************************************************/
bool
checkUser(char *username)
{
  FILE *userfile;
  bool found=false;
  char line[255];
  char *user;

  userfile = fopen(USERFILE, "r");
  if(userfile == NULL)
    return false;

  while( (!feof(userfile)) && !found)
    {
      fscanf(userfile, "%s", line);
      user = strtok(line, ":");
      if(user != NULL)
	if(strcmp(user, username) == 0)
	  found = true;
    }

  fclose(userfile);

  return found;
}

/******************************************************************************
Do the actual login
 *****************************************************************************/
void
login(char *contentString)
{
  char username[255];
  char password[255];
  bool ok=true;
  FILE *datafile;
  char tmpstr[255];
  char room[255];

  if(!GetVar(contentString, "Username", username)) ok=false;
  if(!GetVar(contentString, "Password", password)) ok=false;
  if(!GetVar(contentString, "Room", room)) ok=false;

  if(!ok)
    {
      printf("<HEAD><TITLE>Webchat %s: login</TITLE></HEAD>\n", VERSION);
      printf("<BODY>\n");
      printf("<H1>ERROR: All 3 fields have to be filled</H1>\n");
      printf("Please type both your username and password, and select a chatromm!<BR>\n");
      printf("(Press 'back' in your browser and try again)<BR>\n");
      return;
    }

  if(!checkUserPass(false, username, password, password))
    {
      printf("<HEAD><TITLE>Webchat %s: login</TITLE></HEAD>\n", VERSION);
      printf("<BODY>");
      printf("<H1>ERROR: Wrong username/password</H1>\n");
      printf("(Press 'back' in your browser and try again)");
      return;
    }

  srand(time(NULL));

  strcpy(tmpstr, DATADIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, username);

  datafile = fopen(tmpstr, "r");
  if(datafile != NULL)
    {
      printf("<HEAD><TITLE>Webchat %s: login</TITLE></HEAD>\n", VERSION);
      printf("<BODY>\n");
      printf("<H1>ERROR: User allready logged in</H1>\n");
      printf("If you were disconnected from the server, wait 1 minute and try again!<BR>\n");
      fclose(datafile);
      return;
    }

  datafile = fopen(tmpstr, "w");
  fprintf(datafile, "%i\n", time(NULL));
  fprintf(datafile, "%s\n", room);
  fclose(datafile);

  strcpy(tmpstr, ROOMDIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, room);
  strcat(tmpstr, ".list");

  datafile = fopen(tmpstr, "a");
  
  fprintf(datafile, "%s\n", username);  

  fclose(datafile);

  sprintf(tmpstr, ">> '%s' has joined us", username);
  putMessage(room, tmpstr);

  ConvertString(room);

  printf("<HEAD><TITLE>Webchat %s: %s</TITLE></HEAD>\n", VERSION, room);
  printf("<FRAMESET rows=*,150>\n");
  printf("<FRAME src=\"%s/messages?Username=%s&Key=%s\">\n", scriptName, username, password);
  printf("<FRAME src=\"%s/send?Username=%s&Key=%s\">\n", scriptName, username, password);
  printf("</FRAMESET>");
}

/******************************************************************************
Checks the users username, and password
 *****************************************************************************/
bool
checkUserPass(bool crypted, char *username, char *password, char *retPass)
{
  FILE *userfile;
  char *user;
  char *pass;
  bool matched=false;
  bool found=false;
  char line[255];
  char salt[3];

  salt[2] = 0;
  
  userfile = fopen(USERFILE, "r");
  if(userfile == NULL)
    return false;

  while( !feof(userfile) && !found )
    {
      fscanf(userfile, "%s", line);
      user = strtok(line, ":");
      pass = strtok(NULL, ":");
      if(strcmp(user, username) == 0)
	{
	  char *cpass;
	  found = true;
	  salt[0] = pass[0];
	  salt[1] = pass[1];
	  if(crypted)
	    cpass = password;
	  else
	    cpass = crypt(password, salt);
	  if(strcmp(pass, cpass) == 0)
	    {
	      matched = true;
	      if(retPass != NULL)
		strcpy(retPass, cpass);
	    }
	}
    }

  return matched;
}

/******************************************************************************
The message frame
 *****************************************************************************/
void
showMessages(char *queryString)
{
  FILE *datafile;
  char tmpstr[255];
  char username[255];
  char roomname[255];
  char password[255];

  if(!GetVar(queryString, "Username", username))
    {
      printf("INTERNAL SCRIPT ERROR!");
      return;
    }

  if(!GetVar(queryString, "Key", password))
    {
      printf("Something is wrong here.... this looks like an illegal access");
      return;
    }

  if(!checkUserPass(true, username, password, NULL))
    {
      printf("Something is wrong here.... this looks like an illegal access");
      return;
    }

  strcpy(tmpstr, DATADIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, username);

  datafile = fopen(tmpstr, "r");
  if(datafile == NULL)
    {
      printf("<HEAD><TITLE>Webchat %s: logged out</TITLE></HEAD>", VERSION);
      printf("<BODY>");
      printf("<H1>User '%s' logged out!</H1>\n", username);
      printf("You can return to the <A href=\"%s\">start page</A>, if you want to log into another room<BR>\n", scriptName);
      return;
    }

  printf("<META http-equiv=\"refresh\" content=\"5; url=%s/messages?Username=%s&Key=%s\">", scriptName, username, password);

  printf("<BODY>\n");  

  readLine(datafile, tmpstr);

  readLine(datafile, roomname);

  cleanUsers(roomname);

  fclose(datafile);

  strcpy(tmpstr, ROOMDIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, roomname);
  strcat(tmpstr, ".mesg");

  datafile = fopen(tmpstr, "r");

  if(datafile != NULL)
    {
      while(!feof(datafile))
	{
	  readLine(datafile, tmpstr);
	  if(!feof(datafile))
	    printf("%s<BR>\n", tmpstr);
	}
    
      fclose(datafile);
    }

  strcpy(tmpstr, DATADIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, username);

  datafile = fopen(tmpstr, "w");

  fprintf(datafile, "%i\n", time(NULL));
  fprintf(datafile, "%s\n", roomname);

  fclose(datafile);
}

/******************************************************************************
Read a line from a file
 *****************************************************************************/
void
readLine(FILE *file, char *output)
{
   int ch=-1;
   int count=0;
   char tempStr[255];

   while( (ch!=0x0A) && (!feof(file)) )
   {
      ch = fgetc(file);
      output[count] = ch;
      count++;
   }
   output[count-1] = '\0';
   if(output[count-2] == '\n')
     output[count-2] = '\0';
}

/******************************************************************************
The "Send Message" frame
 *****************************************************************************/
void
sendMessages(char *queryString)
{
  char message[255];
  char username[255];
  char password[255];
  char room[255];
  FILE *datafile;
  char tmpstr[255];

  GetVar(queryString, "Username", username);
  GetVar(queryString, "Key", password);

  if(password[strlen(password)] == ' ')
    password[strlen(password)-1] = 0;

  if(!checkUserPass(true, username, password, NULL))
    {
      printf("This looks like an illegal access(1)<BR>");
      return;
    }

  strcpy(tmpstr, DATADIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, username);

  datafile = fopen(tmpstr, "r");
  if(datafile == NULL)
    {
      printf("USER NOT LOGGED IN!");
      return;
    }

  readLine(datafile, tmpstr);
  readLine(datafile, room);

  fclose(datafile);

  if(GetVar(queryString, "Message", message))
    {
      sprintf(tmpstr, "[%s]: %s", username, message);
      putMessage(room, tmpstr);
    }

  ConvertString(room);

  printf("<BODY>\n");
  printf("<CENTER>\n");
  printf("Room: '%s'<BR>\n", room);
  printf("<FORM action=\"%s/send\" method=get>\n", scriptName);
  printf("<INPUT type=hidden name=Username value=\"%s\">", username);
  printf("<INPUT type=hidden name=Key value=\"%s\">", password);
  printf("Message: <INPUT type=text name=Message width=10>");
  printf("<INPUT type=submit value=Send><BR>\n");
  printf("</FORM>");
  printf("<A href=\"%s/logout?Username=%s&Key=%s\" target=_top>logout</A>\n", scriptName, username, password);

  printf("</CENTER>");
}

/******************************************************************************
Logout a user
 *****************************************************************************/
void
logout(char *username)
{
  FILE *datafile;
  char tmpstr[255];
  char room[255];
  unsigned int fileSize;
  char *data;

  strcpy(tmpstr, DATADIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, username);

  datafile = fopen(tmpstr, "r");
  if(datafile == NULL)
    return;

  readLine(datafile, tmpstr);
  readLine(datafile, room);
  fclose(datafile);

  sprintf(tmpstr, "<< '%s' left us ", username);
  putMessage(room, tmpstr);
  
  strcpy(tmpstr, "rm -f ");
  strcat(tmpstr, DATADIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, username);

  system(tmpstr);

  strcpy(tmpstr, ROOMDIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, room);
  strcat(tmpstr, ".list");


  datafile = fopen(tmpstr, "r");

  for(fileSize = 0; !feof(datafile); fileSize++)
    fgetc(datafile);

  fseek(datafile, 0, SEEK_SET);
  data = (char*) malloc(fileSize);
  
  data[0] = 0;

  while(!feof(datafile))
    {
      readLine(datafile, tmpstr);
      if(strcmp(tmpstr, username) != 0)
	{
	  if(tmpstr[0] != 0)
	    {
	      strcat(data, tmpstr);
	      strcat(data, "\n");
	    }	
	}
    }

  strcpy(tmpstr, ROOMDIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, room);
  strcat(tmpstr, ".list");

  datafile = freopen(tmpstr, "w", datafile);
  fwrite(data, 1, strlen(data), datafile);

  free(data);
  fclose(datafile);
}

/******************************************************************************
Checks the rooms userlist, and sees if anyone hasn't updated the message window
for 40 seconds, and logs them out if they haven't.
 *****************************************************************************/
void
cleanUsers(char *room)
{
  FILE *list;
  char prefix[255];
  char tmpstr[255];
  char name[255];

  strcpy(tmpstr, ROOMDIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, room);
  strcat(tmpstr, ".list");
  list = fopen(tmpstr, "r");

  strcpy(prefix, DATADIR);
  strcat(prefix, "/");

  while(!feof(list))
    {
      FILE *datafile;

      readLine(list, name);

      strcpy(tmpstr, prefix);
      strcat(tmpstr, name);

      if(name[0] != 0)
	{
	  datafile = fopen(tmpstr, "r");
	  
	  if(datafile != NULL)
	    {
	      readLine(datafile, tmpstr);
	      if(time(NULL)-atoi(tmpstr) >= 40)
		{
		  logout(name);
		}
	      fclose(datafile);
	    }
	}    
    }
  
  fclose(list);
}

/******************************************************************************
Puts the message into the rooms messagefile
 *****************************************************************************/
void 
putMessage(char *room, char *message)
{
  FILE *mesgfile;
  unsigned int fileSize;
  char *data;
  int lines;
  char tmpstr[255];

  strcpy(tmpstr, ROOMDIR);
  strcat(tmpstr, "/");
  strcat(tmpstr, room);
  strcat(tmpstr, ".mesg");
  
  mesgfile = fopen(tmpstr, "r");

  ConvertString(message);
  xlate(message);
  
  if(mesgfile != NULL)
    {
      for(fileSize = 0, lines=0; !feof(mesgfile) && lines<NUMLINES-1; fileSize++)
	{
	  if(fgetc(mesgfile) == '\n')
	    lines++;;
	}
      
      fseek(mesgfile, 0, SEEK_SET);
      
      data = (char*) malloc(fileSize);
      fread((char*)data, 1, fileSize, mesgfile);
      
      fclose(mesgfile);
      
      mesgfile = fopen(tmpstr, "w");
      
      fprintf(mesgfile, "%s\n",message);
      fwrite((char*)data, 1, fileSize, mesgfile);
      
      free(data);
      
      fclose(mesgfile);
    }
  else
    {
      mesgfile = fopen(tmpstr, "w");
      
      fprintf(mesgfile, "%s\n", message);
      fclose(mesgfile);
    }

}

/******************************************************************************
Translates the string 'message', by removing all '<','>' and replacing them
with '&lt', '&gt' respectively.
 *****************************************************************************/
void
xlate(char *message)
{
  char *tmpstr;
  unsigned int l=0;

  tmpstr = (char*) malloc(strlen(message));

  for(unsigned int i=0; i<=strlen(message); i++)
    {
      if(message[i] == '<')
	{
	  tmpstr[l] = '&';
	  l++;
	  tmpstr[l] = 'l';
	  l++;
	  tmpstr[l] = 't';
	  l++;
	  tmpstr[l] = ' ';
	  l++;
	}
      else if(message[i] == '>')
	{
	  tmpstr[l] = '&';
	  l++;
	  tmpstr[l] = 'g';
	  l++;
	  tmpstr[l] = 't';
	  l++;
	  tmpstr[l] = ' ';
	  l++;
	}
      else if(message[i] == '&')
	{
	  tmpstr[l] = '&';
	  l++;
	  tmpstr[l] = 'a';
	  l++;
	  tmpstr[l] = 'm';
	  l++;
	  tmpstr[l] = 'p';
	  l++;
	  tmpstr[l] = ' ';
	  l++;
	}
     else if(message[i] == '\"')
	{
	  tmpstr[l] = '&';
	  l++;
	  tmpstr[l] = 'q';
	  l++;
	  tmpstr[l] = 'u';
	  l++;
	  tmpstr[l] = 'o';
	  l++;
	  tmpstr[l] = 't';
	  l++;
	  tmpstr[l] = ' ';
	  l++;
	}	
      else
	{
	  tmpstr[l] = message[i];
	  l++;
	}
    }

  strcpy(message, tmpstr);
  
  free(tmpstr);
}
